#__all__ = ['meshManager']
#import meshManager

#from meshUtil import meshManager


#shManager.py


print('configTools intialized')
#from meshManager.py import MeshManager as ms

#from .Structured import Preprocessor as Structured
#from .Structured2D import Preprocessor as Structured2D
